from __future__ import print_function

import tempfile

from .base import Base
from .hash import HASH_ALGORITHM
from .signature import SIGN_ALGORITHM


class Uploader(Base):
    def upload(self):
        if not self.rsa_key.has_private():
            raise Exception(
                'Private key is needed to sign manifest of uploaded files')

        with tempfile.TemporaryFile(
                prefix='sumsy', suffix='.manifest') as manifest:

            manifest.write('{}\n'.format(HASH_ALGORITHM).encode('utf-8'))

            # Calculate hashes, upload files
            for key, mode, reader in self.file_entries():
                manifest.write(
                    '{} 0{:o} {}\n'.format(
                        self.get_file_hash(reader).hexdigest(), mode, key)
                    .encode('utf-8'))

                print('Uploading {}'.format(key))
                self.upload_file(key, reader)

            # Sign manifest
            manifest_hash = self.get_file_hash(manifest)
            signature = self.sign(manifest_hash)

            manifest.write(
                '---\n'
                '{}\n'
                .format(SIGN_ALGORITHM)
                .encode('utf-8'))
            manifest.write(signature)

            # Upload manifest and update latest
            print('Uploading manifest')
            self.upload_manifest(manifest)

    run = upload



import base64
import boto3
import errno
import os
import tempfile

from .hash import Hash
from .signature import Signature


__metaclass__ = type


def _read_in_chunks(reader, chunk_size=4096):
    while True:
        data = reader.read(chunk_size)
        if not data:
            break
        yield data


class Base:
    def __init__(self, config):
        self.region = config.region

        self.rsa_key = config.rsa_key

        self.bucket = config.bucket
        self.prefix = getattr(config, 'prefix', None)
        self.strip_path_components = getattr(config, 'strip', 0)
        self.manifest_key = config.manifest

        self.cache = getattr(config, 'cache', None)
        self.directory = config.directory

        if not self.directory.endswith(os.path.sep):
            self.directory += os.path.sep

        self._created_directories = set()

        self.s3 = boto3.client('s3', region_name=self.region)

    def file_entries(self):
        directory = os.path.abspath(self.directory) + os.path.sep
        for subdir, _, filenames in os.walk(directory):
            for filename in filenames:
                actual_path = os.path.join(subdir, filename)
                unrooted_path = actual_path[len(directory):]

                # Normalize separators
                if os.path.sep != '/':
                    unrooted_path = unrooted_path.replace(os.path.sep, '/')

                key = self.prefix + '/' + unrooted_path
                with open(actual_path, 'rb') as reader:
                    mode = os.fstat(reader.fileno()).st_mode & 0o777
                    yield key, mode, reader

    def sign(self, mhash):
        return base64.encodestring(Signature(self.rsa_key).sign(mhash))

    def verify(self, mhash, signature):
        return Signature(self.rsa_key).verify(
            mhash, base64.b64decode(signature))

    def get_hash(self, lines):
        h = Hash()
        for line in lines:
            h.update(line)
        return h

    def get_file_hash(self, reader):
        reader.seek(0)

        h = Hash()
        for chunk in _read_in_chunks(reader):
            h.update(chunk)
        return h

    def upload_file(self, key, reader):
        reader.seek(0)
        self.s3.upload_fileobj(reader, self.bucket, key)

    def upload_manifest(self, reader, update_latest=True):
        key = '{}.manifest'.format(self.prefix)
        self.upload_file(key, reader)

        self.s3.copy_object(
            Bucket=self.bucket,
            CopySource=dict(Bucket=self.bucket, Key=key),
            Key=self.manifest_key)

    def download_file(self, key, mode, content_hash):
        with tempfile.NamedTemporaryFile(dir=self.directory) as provisional:
            self.s3.download_fileobj(self.bucket, key, provisional)
            if content_hash != self.get_file_hash(provisional).hexdigest():
                raise Exception(
                    'File hash in manifest does not match downloaded file: {}'
                    .format(key))

            directory, _, filename = key.rpartition('/')

            for i in range(self.strip_path_components):
                directory = directory.partition('/')[2]

            if os.path.sep != '/':
                directory = directory.replace('/', os.path.sep)
            directory = os.path.join(self.directory, directory)

            self.ensure_directory(directory)

            os.fchmod(provisional.fileno(), mode)
            os.rename(provisional.name, os.path.join(directory, filename))

            try:
                provisional.close()
            except OSError as e:
                if e.errno != errno.ENOENT:
                    raise

    def download_manifest(self, writer):
        self.s3.download_fileobj(self.bucket, self.manifest_key, writer)

    def ensure_directory(self, directory):
        if directory in self._created_directories:
            return

        try:
            os.makedirs(directory)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise

        self._created_directories.add(directory)



import os

from Crypto.PublicKey import RSA


__metaclass__ = type


def _write_file(filename, content, mode):
    fd = os.open(filename, os.O_WRONLY | os.O_CREAT | os.O_EXCL, mode)
    with os.fdopen(fd, 'w') as file:
        file.write(content)


class KeyGenerator:
    def __init__(self, config):
        self.key_name = config.key_name
        self.key_size = config.key_size

    def generate_key(self):
        key = RSA.generate(self.key_size)

        _write_file(self.key_name, key.exportKey(), 0o600)
        _write_file(self.key_name + '.pub', key.publickey().exportKey(), 0o644)

    run = generate_key



import argparse

from Crypto.PublicKey import RSA

from .uploader import Uploader
from .downloader import Downloader
from .keygenerator import KeyGenerator


def RSAKeyType(path):
    with open(path, 'rb') as raw_key:
        return RSA.importKey(raw_key.read())


def parse_args():
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers()

    parser.add_argument(
        '--region', default='eu-west-1', help='''AWS region (default:
        eu-west-1)''')


    upload = subparsers.add_parser('upload')
    upload.set_defaults(cls=Uploader)

    upload.add_argument(
        '--private-key', dest='rsa_key', type=RSAKeyType, required=True,
        help='''Private RSA key to use to sign the manifest''')

    upload.add_argument(
        '--bucket', required=True, help='''S3 bucket to upload files and
        manifest to''')
    upload.add_argument(
        '--prefix', required=True, help='S3 key prefix for uploaded files')
    upload.add_argument(
        '--manifest', default='latest', help='''S3 key for manifest (default:
        latest)''')

    upload.add_argument(
        'directory', metavar='SOURCE', help='''Directory to upload to S3; the
        local directory structure will be replicated in S3''')


    download = subparsers.add_parser('download')
    download.set_defaults(cls=Downloader)

    download.add_argument(
        '--public-key', dest='rsa_key', type=RSAKeyType, required=True,
        help='''Public RSA key to use to verify integrity of the manifest''')

    download.add_argument(
        '--bucket', required=True, help='''S3 bucket to download files and
        manifest from''')
    download.add_argument(
        '--manifest', default='latest', help='''S3 key for manifest (default:
        latest)''')

    download.add_argument(
        '--strip', default=0, type=int, help='''Number of path components to
        strip from downloaded files (default: 0)''')

    # TODO: caching
    download.add_argument(
        '--cache', default='.sumsy', help='''Directory to use for caching
        downloaded files (default: .sumsy)''')

    download.add_argument(
        'directory', metavar='DESTINATION', help='''Directory to download files
        to; the directory structure in S3 will be replicated locally''')


    generate_key = subparsers.add_parser('generate-key')
    generate_key.set_defaults(cls=KeyGenerator)

    generate_key.add_argument(
        '--key-size', default=4096, type=int, help='''Key length, or size (in
        bits) of the RSA modulus. Must be a multiple of 256, and no smaller than
        1024 (default: 4096)''')
    generate_key.add_argument(
        'key_name', help='''File name for the generated private key; the public
        key will use the same name with an additional '.pub' suffix''')


    config = parser.parse_args()

    if not hasattr(config, 'cls'):
        parser.print_help()
        raise SystemExit()

    return config.cls(config)


def main():
    actor = parse_args()
    actor.run()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
#-*- coding: utf-8 -*-

from setuptools import find_packages, setup


with open('requirements.txt') as requirements_file:
    requirements = [ requirement.partition(' ')[0] for requirement in requirements_file ]

setup(
    name='sumsy',
    description='A forklift who greatly resembles another forklift',
    version='0.5.0',
    author='João Abecasis',
    author_email='joao@abecasis.name',
    url='https://github.com/biochimia/sumsy',
    packages=find_packages(),
    install_requires=requirements,
    entry_points={
        'console_scripts': [
            'sumsy = sumsy.cli:main',
        ],
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Testing',
        'Topic :: System :: Software Distribution',
        'Topic :: Utilities',
    ],
)



from __future__ import print_function

import re
import tempfile

from .base import Base
from .hash import HASH_ALGORITHM
from .signature import SIGN_ALGORITHM


_DIRTY_PATH = re.compile('(?:^|/)(\.\.?)(?:/|$)')


class Downloader(Base):
    def _validate_entry_path(self, path):
        if path.startswith('/'):
            raise Exception(
                'Unexpected absolute entry in manifest: {}'.format(path))

        match = _DIRTY_PATH.search(path)
        if match:
            if match.group(1) == '.':
                raise Exception(
                    'Manifest entry contains redundant \'./\' segments: {}'
                    .format(path))

            raise Exception(
                'Manifest entry contains disallowed \'../\' segments: {}'
                .format(path))

    def _validate_manifest(self, manifest):
        '''Validates the manifest.

        Returns list of (hash, mode, path)-tuples for the file entries in the
        manifest.
        '''
        manifest.seek(0)

        for hash_algorithm in manifest:
            break
        else:
            raise Exception('Manifest is empty')

        if hash_algorithm[:-1].decode() != HASH_ALGORITHM:
            raise Exception(
                'Unexpected hashing algorithm in manifest: {}'
                .format(hash_algorithm[:-1]))

        entries = []
        for entry in manifest:
            if entry == b'---\n':
                break
            entries.append(entry)
        else:
            raise Exception('Manifest is not trusted: missing signature')

        for sign_algorithm in manifest:
            break
        else:
            raise Exception('Manifest is not trusted: missing signature')

        if sign_algorithm[:-1].decode() != SIGN_ALGORITHM:
            raise Exception(
                'Unexpected signing algorithm in manifest: {}'
                .format(sign_algorithm[:-1]))

        manifest_hash = self.get_hash([hash_algorithm] + entries)
        signature = b''.join(list(manifest))

        if not self.verify(manifest_hash, signature):
            raise Exception(
                'Manifest is not trusted: signature verification failed')

        result = []
        for entry in entries:
            content_hash, mode, path = entry[:-1].split(b' ', 2)

            content_hash = content_hash.decode()
            mode = int(mode, 8)
            path = path.decode()

            self._validate_entry_path(path)

            result.append((content_hash, mode, path))

        return result

    def download(self):
        with tempfile.SpooledTemporaryFile(max_size=1024*1024) as manifest:
            self.download_manifest(manifest)
            entries = self._validate_manifest(manifest)

        self.ensure_directory(self.directory)
        for content_hash, mode, path in entries:
            self.download_file(path, mode=mode, content_hash=content_hash)

    run = download
